#!/usr/bin/env python
"""
update one or more binary files using simple commands written in a script

usage: patch_bin [-h] [--script_file SCRIPT_FILE]
                 [--binary_files BINARY_FILES [BINARY_FILES ...]]
                 [--input_values [INPUT_VALUES [INPUT_VALUES ...]]] [--version]
note: to use the crc command, the crcmod package is needed

defining a value or an address in the data buffer:
5: the value 5
-5: the value -5
#5: the value stored in the variable number 5 (variable number from 0 to 255; 256 to 269 are used internally)
-#5: the opposite of the value stored in the variable number 5
@5: the value of the cursor + 5 (cursor initial value is 0)
@-5: the value of the cursor - 5
@#5: the value of the cursor + the value stored in the variable number 5
@-#5: the value of the cursor - the value stored in the variable number 5
I5: the value of the input value 5 (starting at 0, input values are set in the command line).
-I5: the opposite of the value input value 5
@I5: the value of the cursor + the value of input value 5
@-I5: the value of the cursor - the value of input value 5
SIZE: variable set to the length of the data buffer in unsigned byte
SIZE2: variable set to the length of the data buffer in unsigned short (rounded to inferior)
SIZE4: variable set to the length of the data buffer in unsigned int (rounded to inferior)
TRUE: constant set to 1
FALSE: constant set to 0

commands:
little_endian: set little endian encoding (default)
big_endian: set big endian encoding

cursor [address_value]: set the cursor to the specified address
store [value] #variable_number: store a value in the selected variable
read [address_value] #variable_number: read the data buffer at the specified address
    and store the value in the selected variable
read2 [address_value] #variable_number: same as read but for 2 bytes using current endianness
read4 [address_value] #variable_number: same as read but for 4 bytes using current endianness
write [address_value] [value]: write the specified value at the given address in the data buffer
write2 [address_value] [value]: same as write but for 2 bytes using current endianness
write4 [address_value] [value]: same as write but for 4 bytes using current endianness
swap [address_value1] [address_value2]: swap the values of the data buffer at the specified addresses

insert [adress_value] [length] [value]: insert length bytes with value set to "value" at address address_value
delete [adress_value] [length]: delete length bytes starting at address address_value
import [source_file_name] [src_address] [length] [dst_address] [insert_flag]: import length bytes from src_addres
    in source_file_name to dst_address. if insert_flag is TRUE bytes are inserted else they are overwritten
export [src_address] [length] [destination_file_name]: export length bytes from src_address to a file

loop [number_of_loops] #variable_number: loop on the commands between this one and the end_loop command and
    store the current index starting from 0 into the selected variable (loops can be nested)
end_loop: indicate the end of the loop
if [value]: execute following commands if value > 0. go to "else" if defined or "end_if"
else: execute following commands when the "if" failed
end_if: end of if/else block

macro [macro_name]: define a macro block. macros must be defined at the end of the script
end_macro: end of macro block
call [macro_name]: execute commands of a macro. no recursion allowed

xor [value1] [value2] #variable_number: calculate value1 xor value2 and store the result in the variable variable_number
and [value1] [value2] #variable_number: calculate value1 and value2 and store the result in the variable variable_number
or [value1] [value2] #variable_number: calculate value1 or value2 and store the result in the variable variable_number
note: xor is the bitwise operator. and/or are the logical operators
sum [value1] [value2] #variable_number: calculate value1 + value2 and store the result in the variable variable_number
multiply [value1] [value2] #variable_number: calculate value1 * value2 and store the result in the variable variable_number

equal [value1] [value2] #variable_number: if value1 = value2 the variable is set to TRUE else to FALSE
different [value1] [value2] #variable_number: if value1 != value2 the variable is set to TRUE else to FALSE
inferior [value1] [value2] #variable_number: if value1 < value2 the variable is set to TRUE else to FALSE
greater [value1] [value2] #variable_number: if value1 > value2 the variable is set to TRUE else to FALSE

crc [polynomial] [start_crc] [xor_out] [reverse_on] [address_value] [length] #variable_number: calculate the crc
    from address_value to address_value + length and store the result in the variable variable_number
    if reverse_on is set to TRUE, the reverse algorithm is used

sum, multiply and swap are "macro commands" that will be expanded automatically using elementary commands

sample:
#add 1 to each even byte of the binary file
cursor 0
loop SIZE2 #1
    read @0 #0
    sum #0 1 #0
    write @0 #0
    cursor @2
end_loop

Copyright 2012, David Gemelli

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "David Gemelli"
__version__ = "1.0.0"

import logging
import sys
if sys.version_info < (2, 7):
    import optparse
else:
    import argparse
import os
import struct
try:
    import crcmod
except:
    print("no crc command available. to use it, install crcmod package")

# in python 3, range replaces xrange
if sys.version_info < (3,):
    range = xrange

# default endianness: little endian
endianness = '<'


# dictionary of simple commands
# key: name
# value: number of elements of the command
simple_commands = {
    "read": 3,
    "read2": 3,
    "read4": 3,
    "write": 3,
    "write2": 3,
    "write4": 3,
    "equal": 4,
    "different": 4,
    "greater": 4,
    "inferior": 4,
    "crc": 8,
    "xor": 4,
    "and": 4,
    "or": 4,
    "little_endian": 1,
    "big_endian": 1,
    "cursor": 2,
    "store": 3,
    "insert": 4,
    "delete": 3,
    "import": 6,
    "export": 4,
}


def analyze_line(splits, num_inst, rscript):
    """
    Parse line elements and build script list
    @param splits: list of elements of the script line
    @param num_inst: current instruction number (starting from 0)
    @param rscript: script list to fill
    @return: new instruction number
    """
    rcmd = None

    # one line command ?
    if splits[0] in simple_commands:
        # check number of parameters
        if len(splits) < simple_commands[splits[0]]:
            raise PatchBinError( \
                "wrong number of parameters for %s command" % (splits[0]))
        # build command
        rcmd = [[]]
        for elem in range(simple_commands[splits[0]]):
            rcmd[0].append(splits[elem])
        rcmd[0].append(" ".join(splits[simple_commands[splits[0]]:]))
        num_inst += 1
    elif splits[0] == "multiply":
        # [1] value1 string
        # [2] value2 string
        # [3] variable name to store the result
        # expand command
        if len(splits) < 4:
            raise PatchBinError( \
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["store", "@0", "#256", "multiply start", " ".join(splits[4:])]]
        rcmd.append(["store", splits[1], "#257"])
        rcmd.append(["store", splits[2], "#258"])
        rcmd.append(["cursor", "0"])
        rcmd.append(["loop", "#257", "#259", -1])
        rcmd.append(["cursor", "@#258"])
        rcmd.append(["end_loop", num_inst + 4])
        rcmd.append(["store", "@0", splits[3]])
        rcmd.append(["cursor", "#256", "multiply end"])
        num_inst += 9
    elif splits[0] == "sum":
        # [1] value1 string
        # [2] value2 string
        # [3] variable name to store the result
        # expand command
        if len(splits) < 4:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["store", "@0", "#256", "sum start", " ".join(splits[4:])]]
        rcmd.append(["store", splits[2], "#257"])
        rcmd.append(["cursor", splits[1]])
        rcmd.append(["cursor", "@#257"])
        rcmd.append(["store", "@0", splits[3]])
        rcmd.append(["cursor", "#256", "sum end"])
        num_inst += 6
    elif splits[0] == "swap":
        # [1] = address1 string
        # [2] = address2 string
        # expand command
        if len(splits) < 3:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["read", splits[1], "#256", "swap start", " ".join(splits[3:])]]
        rcmd.append(["read", splits[2], "#257"])
        rcmd.append(["write", splits[1], "#257"])
        rcmd.append(["write", splits[2], "#256", "swap end"])
        num_inst += 4
    # loop block
    elif splits[0] == "loop":
        # [1] = number of loops
        # [2] = variable name to store the current loop starting from 0
        if len(splits) < 3:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["loop", splits[1], splits[2], -1, " ".join(splits[3:])]]
        num_inst += 1
    elif splits[0] == "end_loop":
        # write jump @
        nb_loop = 0
        nb_end = 0
        for look_back in range(num_inst - 1, 0, -1):
            if rscript[look_back][0] == "end_loop":
                nb_end = nb_end + 1
            elif rscript[look_back][0] == "loop":
                if nb_loop == nb_end:
                    rcmd = [["end_loop", look_back, " ".join(splits[1:])]]
                    num_inst += 1
                    break
                else:
                    nb_loop += 1
    # if block
    elif splits[0] == "if":
        # [1] = value to test
        if len(splits) < 2:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["if", splits[1], -1, -1, " ".join(splits[2:])]]
        num_inst += 1
    elif splits[0] == "else" or splits[0] == "end_if":
        # write jump @
        nb_if = 0
        nb_end = 0
        for look_back in range(num_inst - 1, 0, -1):
            if rscript[look_back][0] == "end_if":
                nb_end = nb_end + 1
            elif rscript[look_back][0] == "if":
                if nb_if == nb_end:
                    rcmd = [[splits[0], look_back, " ".join(splits[1:])]]
                    # save jump @ in if
                    if splits[0] == "else":
                        rscript[look_back][2] = num_inst
                    else:
                        rscript[look_back][3] = num_inst
                    num_inst += 1
                    break
                else:
                    nb_if += 1
    # macro
    elif splits[0] == "call":
        # [1] = macro name
        if len(splits) < 2:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        rcmd = [["call", splits[1], -1, " ".join(splits[2:])]]
        num_inst += 1
    elif splits[0] == "macro":
        if len(splits) < 2:
            raise PatchBinError(
                "wrong number of parameters for %s command" % (splits[0]))
        # look back for calls to macro
        for look_back in range(num_inst - 1, 0, -1):
            if rscript[look_back][0] == "call" and \
                rscript[look_back][1] == splits[1]:
                rscript[look_back][2] = num_inst
        rcmd = [["macro", splits[1], -1, " ".join(splits[2:])]]
        num_inst += 1
    elif splits[0] == "end_macro":
        # look back for macro beginning
        for look_back in range(num_inst - 1, 0, -1):
            if rscript[look_back][0] == "macro":
                rcmd = [["end_macro", look_back, " ".join(splits[1:])]]
                num_inst += 1
                break

    # append command to script
    if rcmd:
        rscript.extend(rcmd)
    return num_inst


def check_address(addr, data_buffer, inclusive=False):
    """
    check that address is between 0 and the length of the buffer
    raise an exception if address is out of range
    @param addr: address to check
    @param data_buffer: byte buffer
    @param inclusive: if True, last byte of the buffer is accepted
    """
    if inclusive:
        upper_limit = len(data_buffer)
    else:
        upper_limit = len(data_buffer) - 1
    if not (0 <= addr <= upper_limit):
        raise PatchBinError(
            "address %d is out of the data buffer (length %d)" % \
            (addr, len(data_buffer)))


def parse_value(value_str, cursor, variables, inputs, data_buffer):
    """
    analyse value string and get corresponding integer value
    cases are: 5, -5, #5, -#5, @5, @-5, @#5, @-#5
    @param value_str: string representation of a value
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    @return: integer value represented by the string
    """
    rvalue = None
    start = 0
    pos = 0
    neg = 1

    if value_str == "TRUE":
        rvalue = 1
    elif value_str == "FALSE":
        rvalue = 0
    elif value_str == "SIZE":
        rvalue = len(data_buffer)
    elif value_str == "SIZE2":
        rvalue = len(data_buffer) // 2
    elif value_str == "SIZE4":
        rvalue = len(data_buffer) // 4
    else:
        # relative to cursor
        if value_str[0] == "@":
            start = cursor
            pos = 1
        # opposite
        if value_str[pos] == "-":
            neg = -1
            pos += 1
        # variable
        if value_str[pos] == "#":
            rvalue = start + neg * int(variables[int(value_str[pos + 1:], 0)])
        # input value
        elif value_str[pos] == "I":
            rvalue = start + neg * int(inputs[int(value_str[pos + 1:], 0)])
        else:
            rvalue = start + neg * int(value_str[pos:], 0)

    return rvalue


def do_read(cmd, cursor, variables, inputs, data_buffer):
    """
    read one, two or four bytes at the given position in the data buffer
    @param cmd: read[N] command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get address
    addr = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get variable number
    var_num = int(cmd[2][1:])
    # read value
    if cmd[0] == "read":
        check_address(addr, data_buffer)
        variables[var_num] = data_buffer[addr]
        stype = "unsigned byte"
    elif cmd[0] == "read2":
        src_packet = struct.pack('BB', *data_buffer[addr:addr + 2])
        check_address(addr, data_buffer)
        check_address(addr + 1, data_buffer)
        variables[var_num] = int(struct.unpack(endianness + 'H', src_packet)[0])
        stype = "unsigned short"
    elif cmd[0] == "read4":
        src_packet = struct.pack('BBBB', *data_buffer[addr:addr + 4])
        check_address(addr, data_buffer)
        check_address(addr + 3, data_buffer)
        variables[var_num] = int(struct.unpack(endianness + 'I', src_packet)[0])
        stype = "unsigned int"
    logging.info("read %s at address 0x%04X in variable %d: value %d" % \
        (stype, addr, var_num, variables[var_num]))


def do_store(cmd, cursor, variables, inputs, data_buffer):
    """
    store a value in a variable
    @param cmd: store command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get value
    svalue = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get variable number
    var_num = int(cmd[2][1:])
    variables[var_num] = svalue
    logging.info("store value %d in variable %d" % (svalue, var_num))


def do_logical(cmd, cursor, variables, inputs, data_buffer):
    """
    xor, and, or of two values in a variable
    @param cmd: command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get value 1
    value1 = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get value 2
    value2 = parse_value(cmd[2], cursor, variables, inputs, data_buffer)
    # get variable number
    var_num = int(cmd[3][1:])
    if cmd[0] == "xor":
        variables[var_num] = value1 ^ value2
    elif cmd[0] == "or":
        variables[var_num] = value1 or value2
    elif cmd[0] == "and":
        variables[var_num] = value1 and value2
    logging.info("calculate %d %s %d and store the result %d in variable %d" % \
        (value1, cmd[0], value2, variables[var_num], var_num))


def do_check(cmd, cursor, variables, inputs, data_buffer):
    """
    check if a value is equal/different/inferior/superior to another one
    @param cmd: command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get value 1
    value1 = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get value 2
    value2 = parse_value(cmd[2], cursor, variables, inputs, data_buffer)
    # get variable number
    var_num = int(cmd[3][1:])
    if cmd[0] == "equal":
        variables[var_num] = (value1 == value2) * 1
        check_str = "equal to"
    elif cmd[0] == "different":
        variables[var_num] = (value1 != value2) * 1
        check_str = "different than"
    elif cmd[0] == "greater":
        variables[var_num] = (value1 > value2) * 1
        check_str = "greater than"
    elif cmd[0] == "inferior":
        variables[var_num] = (value1 < value2) * 1
        check_str = "inferior to"
    logging.info("check if %d is %s than %d and store the result %d" % \
        (value1, check_str, value2, variables[var_num]) + \
        "in variable %d" % (var_num))


def do_write(cmd, cursor, variables, inputs, data_buffer):
    """
    write one, two or four bytes at the given position in the data buffer
    @param cmd: write[N] command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get address
    addr = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get value
    val = parse_value(cmd[2], cursor, variables, inputs, data_buffer)
    # write value
    if cmd[0] == "write":
        check_address(addr, data_buffer)
        data_buffer[addr] = val & 0x000000FF
        stype = "unsigned byte"
    elif cmd[0] == "write2":
        check_address(addr, data_buffer)
        check_address(addr + 1, data_buffer)
        src_packet = struct.pack(endianness + 'H', val & 0x0000FFFF)
        for offset, elem in enumerate(src_packet):
            try:
                data_buffer[addr + offset] = struct.unpack('B', elem)[0]
            except TypeError:
                data_buffer[addr + offset] = elem
        stype = "unsigned short"
    elif cmd[0] == "write4":
        check_address(addr, data_buffer)
        check_address(addr + 3, data_buffer)
        src_packet = struct.pack(endianness + 'I', val & 0xFFFFFFFF)
        for offset, elem in enumerate(src_packet):
            try:
                data_buffer[addr + offset] = struct.unpack('B', elem)[0]
            except TypeError:
                data_buffer[addr + offset] = elem
        stype = "unsigned int"

    logging.info("write %s %d at address 0x%04X" % (stype, val, addr))


def do_cursor(cmd, cursor, variables, inputs, data_buffer):
    """
    set the data buffer cursor position
    @param cmd: cursor command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    @return: new cursor position
    """
    # get address
    cursor = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    logging.info("set cursor at position %d" % (cursor))
    return cursor


def do_loop(cmd, cursor, variables, inputs, data_buffer):
    """
    start/continue a loop over one or more commands
    @param cmd: loop command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get loop number
    loop_num = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # variable
    var_num = int(cmd[2][1:])
    # first time in the loop
    if cmd[3] == -1:
        logging.info("new loop %d times in variable %d" % (loop_num, var_num))
        cmd[3] = loop_num - 1
        # set index variable to 0
        variables[var_num] = 0
    else:
        logging.info("loop %d time%s in variable %d" % \
            (cmd[3], (cmd[3] > 1) and "s" or "", var_num))
        cmd[3] = cmd[3] - 1
        # increment index variable
        variables[var_num] = variables[var_num] + 1


def do_end_loop(cmd, cursor, variables, script, num_cmd):
    """
    check if the number of loops is reached, return to loop start or continue
    @param cmd: end_loop command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param script: current script
    @param num_cmd: current command number
    @return: updated current command
    """
    ret_cmd = num_cmd
    # variable
    var_num = int(script[cmd[1]][2][1:])
    # check if the loop is finished
    if script[cmd[1]][3] == 0:
        # clear loop so it can be started again (case of nested loops)
        logging.info("loop finished")
        script[cmd[1]][3] = -1
        # last increment of index variable
        variables[var_num] = variables[var_num] + 1
    else:
        # jump to the line before "loop" (because of the +1 done in the caller)
        logging.info("go to loop start")
        ret_cmd = cmd[1] - 1
    return ret_cmd


def do_if(cmd, cursor, variables, inputs, data_buffer, script, num_cmd):
    """
    check the variable. if FALSE and else is defined, jump to it,
    else jump to end_if
    @param cmd: if command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    @param script: current script
    @param num_cmd: current command number
    @return: updated current command
    """
    ret_cmd = num_cmd
    # value to check
    ifvar = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # check if the value is different from 0
    if ifvar > 0:
        # TRUE => next command
        logging.info("if is TRUE")
    else:
        # else defined
        if cmd[2] != -1:
            # jump to the else line
            logging.info("go to else")
            ret_cmd = cmd[2]
        else:
            # jump to the end_if
            logging.info("go to end if")
            ret_cmd = cmd[3]
    return ret_cmd


def do_call(cmd, script, num_cmd):
    """
    start a macro execution
    @param cmd: call macro command with parameters
    @param script: current script
    @param num_cmd: current command number
    @return: updated current command
    """
    # jump to macro beginning
    ret_cmd = cmd[2]
    # save caller position in macro definition
    script[ret_cmd][2] = num_cmd

    logging.info("calling %s macro" % (cmd[1]))
    return ret_cmd


def do_end_macro(cmd, script, num_cmd):
    """
    return to caller position after a macro execution
    @param cmd: end macro command with parameters
    @param script: current script
    @param num_cmd: current command number
    @return: updated current command
    """
    return script[cmd[1]][2]


def do_endianness(cmd):
    """
    select current endianness
    @param cmd: ["big_endian"] or ["little_endian"]
    """
    global endianness

    logging.info("setting endianness to %s" % (cmd[0].replace('_', ' ')))
    if cmd[0] == "little_endian":
        endianness = '<'
    else:
        endianness = '>'


def do_crc(cmd, cursor, variables, inputs, data_buffer):
    """
    calculate crc on selected bytes
    @param cmd: crc command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data
    """
    # get polynomial
    spoly = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    # get start crc
    sstart = parse_value(cmd[2], cursor, variables, inputs, data_buffer)
    # get xor out
    sxor = parse_value(cmd[3], cursor, variables, inputs, data_buffer)
    # get reverse
    srev = parse_value(cmd[4], cursor, variables, inputs, data_buffer)
    # get start address
    saddr = parse_value(cmd[5], cursor, variables, inputs, data_buffer)
    check_address(saddr, data_buffer)
    # get length
    slen = parse_value(cmd[6], cursor, variables, inputs, data_buffer)
    check_address(saddr + slen - 1, data_buffer)
    # get variable number
    var_num = int(cmd[7][1:])

    # create crc function
    crc_fun = crcmod.mkCrcFun(spoly, initCrc=sstart, rev=(srev == 1),
        xorOut=sxor)
    # create buffer
    sbuff = struct.pack('B' * slen, *data_buffer[saddr:saddr + slen])
    # calculate crc
    ocrc = crc_fun(sbuff)
    variables[var_num] = ocrc

    logging.info(
        "calculate crc from 0x%04X to 0x%04X and store %s in variable %d" % \
        (saddr, saddr + slen - 1, hex(ocrc), var_num))


def do_insert(cmd, cursor, variables, inputs, data_buffer):
    """
    insert bytes with a specified value
    @param cmd: insert command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get start address
    saddr = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    check_address(saddr, data_buffer, inclusive=True)
    # get number of bytes to add
    slen = parse_value(cmd[2], cursor, variables, inputs, data_buffer)
    # get default value
    sval = parse_value(cmd[3], cursor, variables, inputs, data_buffer)

    insert_bytes = [sval] * slen
    data_buffer[saddr:saddr] = insert_bytes

    logging.info("insert %d bytes with value 0x%X at address 0x%04X" % \
        (slen, sval, saddr))


def do_delete(cmd, cursor, variables, inputs, data_buffer):
    """
    delete bytes
    @param cmd: delete command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get start address
    saddr = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    check_address(saddr, data_buffer, inclusive=False)
    # get number of bytes to delete
    slen = parse_value(cmd[2], cursor, variables, inputs, data_buffer)

    del data_buffer[saddr:saddr + slen]

    logging.info("delete %d bytes starting at address 0x%04X" % (slen, saddr))


def do_import(cmd, cursor, variables, inputs, data_buffer):
    """
    import bytes from a file
    @param cmd: import command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # read source data
    bin_data = []
    with open(cmd[1], "rb") as bin_file:
        rval = bin_file.read(1)
        while rval:
            sbyte = struct.unpack("B", rval)
            bin_data.append(sbyte[0])
            rval = bin_file.read(1)

    # get source address
    saddr = parse_value(cmd[2], cursor, variables, inputs, bin_data)
    # get number of bytes to import
    slen = parse_value(cmd[3], cursor, variables, inputs, bin_data)
    # get destination address
    daddr = parse_value(cmd[4], cursor, variables, inputs, data_buffer)
    check_address(daddr, data_buffer, inclusive=True)
    # get insertion flag
    do_insert = (cmd[5] == "TRUE")
    if not do_insert:
        check_address(daddr + slen, data_buffer)

    # fill data_buffer
    if do_insert:
        data_buffer[daddr:daddr] = bin_data[saddr:saddr + slen]
        logging.info("import %d bytes from %s/address 0x%04X and insert " % \
            (slen, cmd[1], saddr) + " at address 0x%04X" % (daddr))
    else:
        for pos, src_byte in enumerate(bin_data[saddr:saddr + slen]):
            data_buffer[daddr + pos] = src_byte
        logging.info("import %d bytes from %s/address 0x%04X and write " % \
            (slen, cmd[1], saddr) + " at address 0x%04X" % (daddr))


def do_export(cmd, cursor, variables, inputs, data_buffer):
    """
    export bytes to a file
    @param cmd: export command with parameters
    @param cursor: current cursor value
    @param variables: list of variables
    @param inputs: array of values from command line
    @param data_buffer: binary data to patch
    """
    # get source address
    saddr = parse_value(cmd[1], cursor, variables, inputs, data_buffer)
    check_address(saddr, data_buffer)
    # get number of bytes to export
    slen = parse_value(cmd[2], cursor, variables, inputs, data_buffer)

    # save file
    with open(cmd[3], "wb") as bin_file:
        for dbyte in data_buffer[saddr:saddr + slen]:
            bin_file.write(struct.pack("B", dbyte))

    logging.info("export %d bytes from 0x%04X to %s" % (slen, saddr, cmd[3]))


def apply_patch(data_buffer, script, inputs):
    """
    apply patch script on the data buffer
    @param data_buffer: buffer of bytes from input file
    @param script: list of commands parsed from input script file
    @param inputs: array of values from command line
    """
    num_cmd = 0
    cursor = 0
    variables = [0] * 270
    while True:
        if num_cmd >= len(script):
            break
        cmd = script[num_cmd]
        # macro found: end of script
        if cmd[0] == "macro":
            break
        logging.debug(cmd)
        # parse command
        if cmd[0] == "read" or cmd[0] == "read2" or cmd[0] == "read4":
            do_read(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "write" or cmd[0] == "write2" or cmd[0] == "write4":
            do_write(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "cursor":
            cursor = do_cursor(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "loop":
            do_loop(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "end_loop":
            num_cmd = do_end_loop(cmd, cursor, variables, script, num_cmd)
        elif cmd[0] == "if":
            num_cmd = do_if(cmd, cursor, variables, inputs, data_buffer,
                script, num_cmd)
        elif cmd[0] == "call":
            num_cmd = do_call(cmd, script, num_cmd)
        elif cmd[0] == "end_macro":
            num_cmd = do_end_macro(cmd, script, num_cmd)
        elif cmd[0] == "store":
            do_store(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "little_endian" or cmd[0] == "big_endian":
            do_endianness(cmd)
        elif cmd[0] == "crc":
            do_crc(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "xor" or cmd[0] == "and" or cmd[0] == "or":
            do_logical(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "equal" or cmd[0] == "different" or \
            cmd[0] == "greater" or cmd[0] == "inferior":
            do_check(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "insert":
            do_insert(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "delete":
            do_delete(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "import":
            do_import(cmd, cursor, variables, inputs, data_buffer)
        elif cmd[0] == "export":
            do_export(cmd, cursor, variables, inputs, data_buffer)
        # next command
        num_cmd += 1


def read_script(file_name):
    """
    read script file and build the list of commands from it
    @param file_name: script file name
    @return: script in the form of a list of commands
    """
    rscript = []
    num_inst = 0
    # open script file
    with open(file_name, "r") as script_file:
        line = script_file.readline()
        while line:
            # build commands list from lines
            line = line.strip()
            if line:
                splits = line.split()
                num_inst = analyze_line(splits, num_inst, rscript)
            # next line
            line = script_file.readline()
    return rscript


def patch_file(file_name, script, input_values):
    """
    read a binary file, patch it using the script and save the result
    @param file_name: name of the binary file to patch
    @param script: script to use
    @param input_values: array of values from command line
    """
    bin_data = []
    logging.info("applying patch on %s:" % (file_name))

    with open(file_name, "rb") as bin_file:
        rval = bin_file.read(1)
        while rval:
            sbyte = struct.unpack("B", rval)
            bin_data.append(sbyte[0])
            rval = bin_file.read(1)
    os.rename(file_name, file_name + ".bak")
    # apply script
    apply_patch(bin_data, script, input_values)
    # save new file
    with open(file_name, "wb") as bin_file:
        for dbyte in bin_data:
            bin_file.write(struct.pack("B", dbyte))


class PatchBinError(Exception):
    """simple custom exception"""

    def __init__(self, except_msg):
        """
        setup exception
        @param except_msg: message to display
        """
        self.except_msg = except_msg

    def __str__(self):
        """
        text representation
        return: exception message
        """
        return self.except_msg


def init_logging():
    logging.basicConfig(level=logging.DEBUG,
        format='%(asctime)s %(levelname)-8s %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S', filename='patch.log',
        filemode='w')
    # INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)


if __name__ == "__main__":

    def int_from_string(input_str):
        """convert string to int"""
        return int(input_str, 0)

    # parse argmuents
    if sys.version_info < (2, 7):
        parser = \
            optparse.OptionParser(usage="usage: %prog <options>",
            version="%prog " + __version__)
        parser.add_option('--script_file', action="store",
            dest="script_file", help='patch script file')
        parser.add_option('--binary_files', action="append",
            dest="binary_files", help='binary file(s) to patch')
        parser.add_option('--input_values', action="append",
            dest="input_values", help='input value(s)')
        (options, _) = parser.parse_args()
        args = options
        # load script
        script = read_script(args.script_file)
    else:
        parser = argparse.ArgumentParser(prog="patch_bin",
            description="Patch binary files using scripted commands")
        parser.add_argument('--script_file', nargs=1,
            help='patch script file', required=True)
        parser.add_argument('--binary_files', nargs='+',
            help='binary file(s) to patch', required=True)
        parser.add_argument('--input_values', nargs='*',
            help='input value(s)', type=int_from_string)
        parser.add_argument('--version', action='version',
            version="%(prog)s " + __version__)
        args = parser.parse_args()
        # load script
        script = read_script(args.script_file[0])

    init_logging()
    logging.info("patch script:")
    for cmd in script:
        logging.info(cmd)

    if args.input_values:
        logging.info("input values: %s" % (args.input_values))

    for cur_file in args.binary_files:
        # patch file with script
        patch_file(cur_file, script, args.input_values)
